@@ -28,13 +28,17 @@
 from __future__ import absolute_import
 
 from .util import (
-    click_force_option, click_callback, click_parse_key, click_skip_on_help)
-from ..util import TRANSPORT, generate_static_pw, modhex_decode, modhex_encode
+    click_force_option, click_callback, click_parse_b32_key,
+    click_skip_on_help, prompt_for_touch)
+from ..util import (
+    TRANSPORT, generate_static_pw, modhex_decode,
+    modhex_encode, parse_key, parse_b32_key)
 from binascii import a2b_hex, b2a_hex
 from ..driver_otp import YkpersError
 import os
 import struct
 import click
+import hashlib
 
 
 def parse_hex(length):
@@ -239,65 +243,154 @@ a random one may be generated.
 
     try:
         dev.driver.program_static(slot, password, not no_enter)
-        click.echo('Static password set for slot {}.'.format(slot))
     except YkpersError:
         _failed_to_write_msg(ctx)
 
 
 @slot.command()
 @click_slot_argument
+@click.argument('key', required=False)
 @click.option(
-    '-k', '--key', metavar='HEX', callback=click_parse_key, required=False,
-    help='HMAC-SHA1 secret key.')
-@click.option('--require-touch', is_flag=True, help='Require physical button '
-              'press to generate response.')
+    '-t', '--touch', is_flag=True, help='Require touch'
+    ' on YubiKey to generate response.')
+@click.option(
+        '-T', '--totp', is_flag=True, required=False,
+        help='Use a base32 encoded key for TOTP credentials.')
 @click_force_option
 @click.pass_context
-def chalresp(ctx, slot, key, require_touch, force):
+def chalresp(ctx, slot, key, totp, touch, force):
     """
     Program a challenge-response credential.
 
-    If --key is not given, a randomly generated key will be used.
+    If key is not given, a randomly generated key will be used.
     """
     dev = ctx.obj['dev']
+
     if not key:
-        click.echo('Using a randomly generated key.')
-        key = os.urandom(20)
+        if totp:
+            while True:
+                key = click.prompt('Enter a secret key (base32)')
+                try:
+                    key = parse_b32_key(key)
+                    break
+                except Exception as e:
+                    click.echo(e)
+                    pass
+        else:
+            key = click.prompt(
+                'Enter a secret key [blank to randomly generate]',
+                default='', show_default=False)
+            if force or key == '':
+                key = os.urandom(20)
+                click.echo('Using a randomly generated key: {}'.format(
+                    b2a_hex(key).decode('ascii')))
+            else:
+                key = parse_key(key)
+    else:
+        if totp:
+            key = parse_b32_key(key)
+            if len(key) > 64:  # Keys longer than 64 bytes are hashed.
+                key = hashlib.sha1(key).digest()
+            if len(key) > 20:
+                ctx.fail('YubiKey Slots cannot handle TOTP keys over 20 bytes.')
+            key += b'\x00' * (20 - len(key))  # Keys must be padded to 20 bytes.
+        else:
+            key = parse_key(key)
 
-    force or click.confirm('Program a challenge-response credential in slot {}?'
-                           .format(slot), abort=True)
-
-    click.echo('Programming challenge-response in slot {}...'.format(slot))
+    cred_type = 'TOTP' if totp else 'challenge-response'
+    force or click.confirm('Program a {} credential in slot {}?'
+                           .format(cred_type, slot), abort=True)
     try:
-        dev.driver.program_chalresp(slot, key, require_touch)
+        dev.driver.program_chalresp(slot, key, touch)
     except YkpersError:
         _failed_to_write_msg(ctx)
 
 
 @slot.command()
 @click_slot_argument
-@click.argument('key', callback=click_parse_key)
-@click.option('--digits', type=click.Choice(['6', '8']), default='6',
-              callback=lambda c, p, v: int(v),
-              help='Number of digits to output for HOTP codes.')
-@click.option('--imf', type=int, default=0,
-              help='Initial moving factor for credential.')
+@click.argument('challenge', required=False)
+@click.option(
+    '-T', '--totp', is_flag=True, help='Generate a TOTP code, '
+    'use the current time as challenge.')
+@click.option(
+    '-d', '--digits', type=click.Choice(['6', '8']), default='6',
+    help='Number of digits in generated TOTP code (default is 6).')
+@click.pass_context
+def calculate(ctx, slot, challenge, totp, digits):
+    """
+    Perform a challenge-response operation.
+
+    Send a challenge (in hex) to a YubiKey slot with a challenge-response
+credential, and read the response. Supports output as a OATH-TOTP code.
+    """
+    dev = ctx.obj['dev']
+    if not challenge and not totp:
+        ctx.fail('No challenge provided.')
+
+    # Check that slot is not empty
+    slot1, slot2 = dev.driver.slot_status
+    if (slot == 1 and not slot1) or (slot == 2 and not slot2):
+        ctx.fail('Cannot perform challenge-response on an empty slot.')
+
+    # Timestamp challenge should be int
+    if challenge and totp:
+        try:
+            challenge = int(challenge)
+        except:
+            ctx.fail('Timestamp challenge for TOTP must be an integer.')
+    try:
+        res = dev.driver.calculate(
+            slot, challenge, totp=totp,
+            digits=int(digits), wait_for_touch=False)
+    except YkpersError as e:
+        # Touch is set
+        if e.errno == 11:
+            prompt_for_touch()
+            try:
+                res = dev.driver.calculate(
+                    slot, challenge, totp=totp,
+                    digits=int(digits), wait_for_touch=True)
+            except YkpersError as e:
+                # Touch timed out
+                if e.errno == 4:
+                    ctx.fail('The YubiKey timed out.')
+        else:
+            ctx.fail('Failed to calculate challenge.')
+    click.echo(res)
+
+
+@slot.command()
+@click_slot_argument
+@click.argument('key', callback=click_parse_b32_key, required=False)
+@click.option('-d', '--digits', type=click.Choice(['6', '8']), default='6',
+              help='Number of digits in generated code (default is 6).')
+@click.option('-c', '--counter', type=int, default=0,
+              help='Initial counter value.')
 @click.option('--no-enter', is_flag=True, help="Don't send an Enter "
-              'keystroke after outputting an OTP.')
+              'keystroke after outputting the code.')
 @click_force_option
 @click.pass_context
-def hotp(ctx, slot, key, digits, imf, no_enter, force):
+def hotp(ctx, slot, key, digits, counter, no_enter, force):
     """
     Program an HMAC-SHA1 OATH-HOTP credential.
 
-    KEY is given as a hex or base32 encoded string.
     """
     dev = ctx.obj['dev']
-    force or click.confirm('Program a HOTP credential in slot {}?'.format(slot),
-                           abort=True)
-    click.echo('Programming HOTP credential in slot {}...'.format(slot))
+    if not key:
+        while True:
+            key = click.prompt('Enter a secret key (base32)')
+            try:
+                key = parse_b32_key(key)
+                break
+            except Exception as e:
+                click.echo(e)
+                pass
+
+    force or click.confirm(
+        'Program a HOTP credential in slot {}?'.format(slot), abort=True)
     try:
-        dev.driver.program_hotp(slot, key, imf, digits == 8, not no_enter)
+        dev.driver.program_hotp(
+            slot, key, counter, int(digits) == 8, not no_enter)
     except YkpersError:
         _failed_to_write_msg(ctx)
 
