@@ -31,11 +31,12 @@ import hashlib
 import struct
 import hmac
 import six
+import time
 from functools import total_ordering
 from enum import IntEnum
 from .driver_ccid import APDUError, OATH_AID, SW_OK
 from .util import (
-    Tlv, parse_tlvs, time_challenge, derive_key,
+    Tlv, parse_tlvs, time_challenge,
     format_code, parse_truncated, hmac_shorten_key)
 
 
@@ -96,7 +97,9 @@ class SW(IntEnum):
 @total_ordering
 class Credential(object):
 
-    def __init__(self, name, code=None, oath_type='', touch=False, algo=None):
+    def __init__(
+            self, name, code=None, oath_type='', touch=False, algo=None,
+            expiration=None):
         self.name = name
         self.code = code
         self.oath_type = oath_type
@@ -104,10 +107,21 @@ class Credential(object):
         self.algo = algo
         self.hidden = name.startswith('_hidden:')
         self.steam = name.startswith('Steam:')
+        self.expiration = expiration
 
     def __lt__(self, other):
         return self.name.lower() < other.name.lower()
 
+    def to_dict(self):
+        return dict(self.__dict__)
+
+    @staticmethod
+    def from_dict(data):
+        kwargs = dict(data)
+        del kwargs['steam']
+        del kwargs['hidden']
+        return Credential(**kwargs)
+
 
 class OathController(object):
 
@@ -119,9 +133,13 @@ class OathController(object):
     def version(self):
         return self._version
 
+    @property
+    def id(self):
+        return self._id
+
     @property
     def locked(self):
-        return self._challenge
+        return self._challenge is not None
 
     @property
     def _426device(self):
@@ -187,7 +205,7 @@ class OathController(object):
             yield cred
             resp = resp[3 + length:]
 
-    def calculate(self, cred):
+    def calculate(self, cred, timestamp=None):
 
         # The 4.2.0-4.2.6 firmwares have a known issue with credentials that
         # require touch: If this action is performed within 2 seconds of a
@@ -197,8 +215,9 @@ class OathController(object):
         if self._426device and cred.touch:
             self._send_invalid_apdu()
 
-        challenge = time_challenge() \
-            if cred.oath_type == 'totp' else b''
+        if timestamp is None:
+            timestamp = int(time.time())
+        challenge = time_challenge(timestamp)
         data = Tlv(TAG.NAME, cred.name.encode('utf-8')) + \
             Tlv(TAG.CHALLENGE, challenge)
         resp = self.send_apdu(0, INS.CALCULATE, 0, 0x01, data)
@@ -207,14 +226,18 @@ class OathController(object):
         code = resp[1:]
         code = parse_truncated(code)
         cred.code = format_code(code, digits, steam=cred.steam)
+        if cred.oath_type != 'hotp':
+            cred.expiration = ((timestamp + 30) // 30) * 30
         return cred
 
     def delete(self, cred):
         data = Tlv(TAG.NAME, cred.name.encode('utf-8'))
         self.send_apdu(0, INS.DELETE, 0, 0, data)
 
-    def calculate_all(self):
-        data = Tlv(TAG.CHALLENGE, time_challenge())
+    def calculate_all(self, timestamp=None):
+        if timestamp is None:
+            timestamp = int(time.time())
+        data = Tlv(TAG.CHALLENGE, time_challenge(timestamp))
         resp = self.send_apdu(0, INS.CALCULATE_ALL, 0, 0x01, data)
         tlvs = parse_tlvs(resp)
         while tlvs:
@@ -225,18 +248,18 @@ class OathController(object):
             if resp.tag == TAG.TRUNCATED_RESPONSE:
                 cred.oath_type = 'totp'
                 if cred.steam:
-                    cred = self.calculate(cred)
+                    cred = self.calculate(cred, timestamp)
                 else:
                     code = parse_truncated(resp.value[1:])
                     cred.code = format_code(code, digits)
+                    cred.expiration = ((timestamp + 30) // 30) * 30
             elif resp.tag == TAG.HOTP:
                 cred.oath_type = 'hotp'
             elif resp.tag == TAG.TOUCH:
                 cred.touch = True
             yield cred
 
-    def set_password(self, password):
-        key = derive_key(self._id, password)
+    def set_password(self, key):
         keydata = bytearray([OATH_TYPE.TOTP | ALGO.SHA1]) + key
         challenge = os.urandom(8)
         response = hmac.new(key, challenge, hashlib.sha1).digest()
@@ -247,8 +270,7 @@ class OathController(object):
     def clear_password(self):
         self.send_apdu(0, INS.SET_CODE, 0, 0, Tlv(TAG.KEY, b''))
 
-    def validate(self, password):
-        key = derive_key(self._id, password)
+    def validate(self, key):
         response = hmac.new(key, self._challenge, hashlib.sha1).digest()
         challenge = os.urandom(8)
         verification = hmac.new(key, challenge, hashlib.sha1).digest()
