@@ -27,14 +27,20 @@
 
 import os
 import six
-import time
 import struct
 import hashlib
+import re
 from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
 from cryptography.hazmat.primitives import hashes
 from cryptography.hazmat.backends import default_backend
 from enum import IntEnum
+from base64 import b32decode
 from binascii import b2a_hex, a2b_hex
+try:
+    from urlparse import urlparse, parse_qs
+    from urllib import unquote
+except ImportError:
+    from urllib.parse import unquote, urlparse, parse_qs
 
 
 class BitflagEnum(IntEnum):
@@ -239,6 +245,11 @@ def format_code(code, digits=6, steam=False):
         return ('%%0%dd' % digits) % (code % 10 ** digits)
 
 
+def parse_totp_hash(resp):
+    offs = six.indexbytes(resp, -1) & 0xf
+    return parse_truncated(resp[offs:offs+4])
+
+
 def parse_truncated(resp):
     return struct.unpack('>I', resp)[0] & 0x7fffffff
 
@@ -256,7 +267,38 @@ def hmac_shorten_key(key, algo):
     return key
 
 
-def time_challenge(t=None):
-    if t is None:
-        t = time.time()
-    return struct.pack('>q', int(t // 30))
+def time_challenge(timestamp):
+    return struct.pack('>q', int(timestamp // 30))
+
+
+def parse_uri(val):
+    try:
+        uri = val.strip()
+        parsed = urlparse(uri)
+        assert parsed.scheme == 'otpauth'
+        params = dict((k, v[0]) for k, v in parse_qs(parsed.query).items())
+        params['name'] = unquote(parsed.path)[1:]  # Unquote and strip leading /
+        params['type'] = parsed.hostname
+        # Issuer can come both in a param and inside name param.
+        # We store both in the name field on the key.
+        if 'issuer' in params \
+                and not params['name'].startswith(params['issuer']):
+                    params['name'] = params['issuer'] + ':' + params['name']
+        return params
+    except:
+        raise ValueError('URI seems to have the wrong format.')
+
+
+def parse_key(val):
+    val = val.upper()
+    if re.match(r'^([0-9A-F]{2})+$', val):  # hex
+        return a2b_hex(val)
+    else:
+        # Key should be b32 encoded
+        return parse_b32_key(val)
+
+
+def parse_b32_key(key):
+    key = key.upper().replace(' ', '')
+    key += '=' * (-len(key) % 8)  # Support unpadded
+    return b32decode(key)
